# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the masci-tools package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/masci-tools    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit                                        #
###############################################################################
'''
masci-tools
'''
import logging

__version__ = '0.4.3'

logging.getLogger(__name__).addHandler(logging.NullHandler())
